
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/ip_icmp.h>
#include <netdb.h>

void resolve_address(struct sockaddr *addr, char *hostname)
{
	struct sockaddr_in *address;
	struct hostent *host;

	address=(struct sockaddr_in *)addr;
	(void)bzero((char *)address, sizeof(struct sockaddr_in));
	address->sin_family=AF_INET;
	address->sin_addr.s_addr = inet_addr(hostname);

	if((int)address->sin_addr.s_addr==-1) {
		host=gethostbyname(hostname);
		if(host) {
			bcopy(host->h_addr, (char *)&address->sin_addr, host->h_length);
		} else {
			puts("Couldn't resolve address!!!");
			exit(-1);
		}
	}
}

void create_ip(struct iphdr *ip_header, unsigned long source,
		unsigned long dest, u_char protocol, u_char ttl,
		u_short id, int data_len)
{
	ip_header->version = 4;
	ip_header->tos = 0;
	ip_header->frag_off = 0;
	ip_header->check = 0;
	ip_header->saddr = source;
	ip_header->daddr = dest;
	ip_header->protocol = protocol;
	ip_header->ttl = ttl;
	ip_header->id = htons(id);
	ip_header->ihl = 5;
	ip_header->tot_len = htons(sizeof(struct iphdr) + data_len);
}

void create_tcp(struct tcphdr *tcp_header, unsigned long source,
		unsigned long dest, unsigned short sport, unsigned short dport,
		unsigned long seqnum, unsigned long acknum, unsigned char flags)
{
	tcp_header->th_sport=sport;
	tcp_header->th_dport=dport;
	tcp_header->th_seq=seqnum;
	tcp_header->th_ack=acknum;
	tcp_header->th_flags=flags;
	tcp_header->th_sum=0;
}

void create_udp(struct udphdr *udp_header, unsigned long source,
		unsigned long dest, u_short sport, u_short dport, int datalen)
{
	udp_header->source = htons(sport);
	udp_header->dest = htons(dport);
	udp_header->len = htons(sizeof(struct udphdr) + datalen);
	udp_header->check = 0;
}

static struct sockaddr_in local, remote, gw;

void sendip(int sock, char *packet, int mtu, int frag)
{
	static unsigned short id=0;
	struct iphdr *ip=(struct iphdr *)packet;
	struct iphdr ipsv;
	int result;

	memcpy((char *)&ipsv, (char *)ip, sizeof(*ip));

	ip->tot_len=htons(ip->tot_len);
	ip->frag_off=htons(ip->frag_off);

	if(!ip->version)
		ip->version=4;
	if(!ip->id)
		ip->id=htons(id++);
	if(!ip->ttl)
		ip->ttl=60;

	if(!frag || (ntohs(ip->tot_len) < mtu)) {
		ip->check=0;
		ip->check=chksum((unsigned short *)ip, ip->ihl<<2);

		result = sendto(sock, packet, ntohs(ip->tot_len), 0,
			    (struct sockaddr *) &remote, sizeof(remote));
		if (result != ntohs(ip->tot_len)) {
			perror("sending packet");
		}
	} else {
		char optcpy[48],ol;
		char *s;
		int i,iplen,sent=0,ts,hlen,olen;

		hlen=ip->ihl<<2;
		if(mtu<(hlen+8)) {
			fprintf(stderr, "mtu (%d) < ip header size (%d) + 8\n", mtu, hlen);
			fprintf(stderr, "Can't fragment data\n");
			return;
		}

		ol=(ip->ihl<<2)-sizeof(*ip);
		for(i=0,s=(char *)(ip+1); ol>0;)
			if(*s==IPOPT_END) {
				optcpy[i++]=*s;
				break;
			} else if(*s==IPOPT_NOOP) {
				s++;
				ol--;
			} else {
				olen=(int)(*(unsigned char *)(s+1));
				ol-=olen;
				if((*s)&0x80) {
					memcpy(optcpy+i,s,olen);
					i+=olen;
					s+=olen;
				}
			}

		if(i) {
			while((i&3)&&(i&3)!=3)
				optcpy[i++]=IPOPT_NOOP;
			if(i&3==3)
				optcpy[i++]=IPOPT_END;
		}

		s=(char *)ip+hlen;
		iplen=ntohs(ip->tot_len)-hlen;
		ip->frag_off|=htons(0x2000);

		while(1) {
			char ipbuf[65536];

			if((sent+(mtu-hlen)) >= iplen) {
				ip->frag_off ^= htons(0x2000);
				ts=iplen-sent;
			} else
				ts=(mtu-hlen);

			ip->frag_off &= htons(0xe000);
			ip->frag_off |= htons(sent >> 3);
			ts+=hlen;
			ip->tot_len=htons(ts);
			ip->check=0;
			ip->check=chksum((unsigned short *)ip, hlen);

			memcpy(ipbuf, (char *)ip, hlen);
			memcpy(ipbuf+hlen, s+sent, ts-hlen);

			result = sendto(sock, ipbuf, ts, 0,
				    (struct sockaddr *) &remote, sizeof(remote));
			if (result != ts) {
				perror("sending packet");
			}

			sent += (ts-hlen);
			if(!(ntohs(ip->frag_off) & 0x2000))
				break;
			else if(!(ip->frag_off & htons(0x1fff))) {
				hlen=i+sizeof(*ip);
				ip->ihl=(sizeof(*ip)+i)>2;
				memcpy((char *)(ip+1), optcpy, i);
			}
		}
	}

	memcpy((char *)ip, (char *)&ipsv, sizeof(*ip));
}

void sendraw(int sock, char *packet)
{
	struct iphdr *ip=(struct iphdr *)packet;
	int result;

	result = sendto(sock, packet, ntohs(ip->tot_len), 0,
		    (struct sockaddr *) &remote, sizeof(remote));
	if (result != ntohs(ip->tot_len)) {
		perror("sending packet");
	}
}

struct phdr {
	__u32 saddr;
	__u32 daddr;
	__u8 zero;
	__u8 protocol;
	__u16 len;
};

void sendtcp(int sock, char *packet, int mtu)
{
	struct iphdr *ip=(struct iphdr *)packet;
	struct tcphdr *t=(struct tcphdr *)((char *)ip+(ip->ihl <<2));
	struct tcphdr tsv;
	struct phdr *ph;
	unsigned char lbuf[32];
	int thlen;
	static tcp_seq iss=2;

	memcpy((char *)&tsv, (char *)t, sizeof(*t));

	ip->protocol=IPPROTO_TCP;

	ph=(struct phdr *)lbuf;
	bzero((char *)ph, sizeof(*ph));
	thlen=sizeof(struct tcphdr);

	t->th_sport=htons(t->th_sport);
	t->th_dport=htons(t->th_dport);
	t->th_sum=0;
	
	ph->protocol=ip->protocol;
	ph->saddr=ip->saddr;
	ph->daddr=ip->daddr;

	if(!t->th_win)
		t->th_win=htons(4096);
	if(!t->th_seq)
		t->th_seq=htonl(iss);
	iss+=64;

	if((t->th_flags==TH_SYN) && !ip->frag_off) {
		t->th_ack=htonl(0x020405b4);
	}

	if(!t->th_off)
		t->th_off=thlen>>2;
	else
		thlen=t->th_off<<2;
		
	ph->len=htons(thlen);
	ip->tot_len=(ip->ihl<<2)+thlen;

	memcpy(lbuf+sizeof(struct phdr), (char *)ip+(ip->ihl<<2), thlen);

	t->th_sum=chksum((unsigned short *)lbuf, thlen+sizeof(struct phdr));

	sendip(sock, packet, mtu, 1);

	memcpy((char *)t, (char *)&tsv, sizeof(*t));
}

void sendudp(int sock, char *packet, int mtu)
{
	struct iphdr *ip=(struct iphdr *)packet;
	struct udphdr *u=(struct udphdr *)((char *)ip+(ip->ihl <<2));
	struct udphdr usv;
	struct phdr *ph;
	unsigned char lbuf[32];
	int hlen;

	memcpy((char *)&usv, (char *)u, sizeof(*u));

	ph=(struct phdr *)lbuf;
	bzero((char *)ph, sizeof(*ph));

	u->source=htons(u->source);
	u->dest=htons(u->dest);
	u->len=htons(u->len);
	u->check=0;
	
	ph->protocol=ip->protocol;
	ph->saddr=ip->saddr;
	ph->daddr=ip->daddr;
	ph->len=u->len;

	memcpy(lbuf+sizeof(struct phdr), (char *)ip+(ip->ihl<<2), ntohs(u->len));

	u->check=chksum((unsigned short *)lbuf, ntohs(u->len)+sizeof(struct phdr));

	sendip(sock, packet, mtu, 1);

	memcpy((char *)u, (char *)&usv, sizeof(*u));
}

void sendicmp(int sock, char *packet, int mtu)
{
	struct iphdr *ip=(struct iphdr *)packet;
	struct icmphdr *ic;

	ic=(struct icmphdr *)((char *)ip+(ip->ihl<<2));
	ic->checksum=0;
	ic->checksum=chksum((unsigned short *)ic, sizeof(struct icmphdr));

	return(sendip(sock, packet, mtu, 1));
}

extern	char	*optarg;
extern	int	optind;

char	options[68];
char	default_device[] = "eth0";

void	usage(prog)
char	*prog;
{
	fprintf(stderr, "Usage: %s [options] dest\n\
\toptions:\n\
\t\t-d device\tSend out on this device\n\
\t\t-g gateway\tIP gateway to use if non-local dest.\n\
\t\t-m mtu\t\tfake MTU to use when sending out\n\
\t\t-p pointtest\t\n\
\t\t-s src\t\tsource address for IP packet\n\
\t\t-1 \t\tPerform test 1 (IP header)\n\
\t\t-2 \t\tPerform test 2 (IP options)\n\
\t\t-3 \t\tPerform test 3 (ICMP)\n\
\t\t-4 \t\tPerform test 4 (UDP)\n\
\t\t-5 \t\tPerform test 5 (TCP)\n\
\t\t-6 \t\tPerform test 6 (overlapping fragments)\n\
\t\t-7 \t\tPerform test 7 (random packets)\n\
\t\t-8 \t\tPerform test 8 (missing UDP header)\n\
", prog);
	exit(1);
}


main(argc, argv)
int	argc;
char	**argv;
{
	char packet[65536];
	int sock = 0;
	struct	in_addr	gwip;
	char	*name =  argv[0], host[64], *gateway = NULL, *dev = NULL;
	char	*src = NULL, *dst, c, *s;
	int	mtu = 1500, tests = 0, pointtest = 0;
	int true=1;

	while ((c = getopt(argc, argv, "12345678IP:TUd:f:g:m:o:p:s:t:")) != -1)
		switch (c)
		{
		case '1' :
		case '2' :
		case '3' :
		case '4' :
		case '5' :
		case '6' :
		case '7' :
		case '8' :
			tests = c - '0';
			break;
		case 'd' :
			dev = optarg;
			break;
		case 'g' :
			gateway = optarg;
			break;
		case 'm' :
			mtu = atoi(optarg);
			if (mtu < 28)
			    {
				fprintf(stderr, "mtu must be > 28\n");
				exit(1);
			    }
			break;
		case 'p' :
			pointtest = atoi(optarg);
			break;
		case 's' :
			src = optarg;
			break;
		default :
			fprintf(stderr, "Unknown option \"%c\"\n", c);
			usage(name);
		}

	if (argc - optind < 2 && !tests)
		usage(name);
	dst = argv[optind++];

	if (!src) {
		gethostname(host, sizeof(host));
		src = host;
	}

	if(!gateway)
		gateway=dst;

	resolve_address((struct sockaddr *)&local, src);
	resolve_address((struct sockaddr *)&remote, dst);
	resolve_address((struct sockaddr *)&gw, gateway);
	sock = socket(AF_INET, SOCK_RAW, 255);
	if (sock == -1) {
		perror("Getting raw socket");
		exit(-1);
	}

	setsockopt(sock, SOL_SOCKET, SO_BROADCAST, &true, sizeof(int));

	if (!dev)
		dev = default_device;
	printf("Device:  %s\n", dev);
	printf("Source:  %s\n", inet_ntoa(local.sin_addr.s_addr));
	printf("Dest:    %s\n", inet_ntoa(remote.sin_addr.s_addr));
	printf("Gateway: %s\n", inet_ntoa(gw.sin_addr.s_addr));
	printf("mtu:     %d\n", mtu);

	create_ip((struct iphdr *)packet, local.sin_addr.s_addr,
		remote.sin_addr.s_addr, 0, 0, 0, 0);
	
	switch(tests) {
	case 1:
		return(ip_test1(sock, packet, mtu, pointtest));
	case 2:
		return(ip_test2(sock, packet, mtu, pointtest));
	case 3:
		return(ip_test3(sock, packet, mtu, pointtest));
	case 4:
		return(ip_test4(sock, packet, mtu, pointtest));
	case 5:
		return(ip_test5(sock, packet, mtu, pointtest));
	case 6:
		return(ip_test6(sock, packet, mtu, pointtest));
	case 7:
		return(ip_test7(sock, packet, mtu, pointtest));
	case 8:
		return(ip_test8(sock, packet, mtu, pointtest));
	default:
		break;
	}
	return 0;
}
